using System;
using System.Collections;

namespace Printy.Library
{
  /// <summary>
  /// Grammar-Knoten, Wurzel eines ATG-Syntaxgraphen
  /// </summary>
  public class Grammar : Node
  {
    /// <summary>
    /// Konstruktor.
    /// </summary>
    /// <param name="text">Name der Grammatik</param>
    internal Grammar(string text) : base(Grammar,text)
    {}

    /// <summary>
    /// Abbildung von Zeichenketten auf Node-Objekte.
    /// Bildet Namen von Produktionen auf die adquaten Produktions-Knoten ab.
    /// Die Liste ist sortiert.
    /// </summary>
    internal readonly SortedList nts = new SortedList();

    /// <summary>
    /// Maximale Zeilenanzahl eines mehrzeiligen Symbols (Kommentar, semantische Aktion,
    /// Attributliste, Konfliktbedingung, Prolog, Epilog).
    /// </summary>
    internal int maxLines;

    /// <summary>
    /// Gibt die Produktion mit dem angegebenem sortieren Index zurck.
    /// D.h. grammar[i] wird typischerweise eine andere Produktion liefern
    /// als grammar.GetByIndex(i).
    /// </summary>
    /// <param name="index"></param>
    /// <returns></returns>
    public Node GetByIndex(int index)
    {
      return GetNode((Node)nts.GetByIndex(index));
    }

    /// <summary>
    /// Gibt die Produktion mit dem angegebenen Namen zurck oder
    /// null, falls es sie nicht gibt.
    /// </summary>
    /// <param name="name"></param>
    /// <returns></returns>
    public Node GetByName(string name)
    {
      return GetNode((Node)nts[name]);
    }


    /// <summary>
    /// Indizierter Zugriff ber den Namen einer Produktion.
    /// Liefer nullt, falls die Produktion nicht existiert.
    /// </summary>
    public Node this[string name]
    {
      get
      {
        return GetNode((Node)nts[name]);
      }
    }
  }
}
