
using System;

namespace Printy.Library {



public class Parser {
	const int _EOF = 0;
	const int _ident = 1;
	const int _string = 2;
	const int _char = 3;
	const int maxT = 24;

	const bool T = true;
	const bool x = false;
	const int minErrDist = 2;
	
	public Scanner scanner;
	public Errors  errors;

	public Token t;    // last recognized token
	public Token la;   // lookahead token
	int errDist = minErrDist;

public Grammar grammar;



	public Parser(Scanner scanner) {
		this.scanner = scanner;
		errors = new Errors();
	}

  /// <summary>
  /// PG 2005-07-04
  /// Registriert ein Kommentar und legt ihn in der Datenstruktur ab.
  /// </summary>
  /// <param name="t">Erkanntes Token</param>
  /// <param name="la">Token des Kommentars</param>
  void RegisterComment(Token t, Token la) 
  {
    RegisterLines(la.lines = scanner.Line - la.line);
    while (t.comment != null)
    {
      t = t.comment;
    }
    t.comment = la;
  }
  
  /// <summary>
  /// PG 2005-07-10
  /// Registriert die Zeilenanzahl eines eventuell mehrzeigigen Symbols und legt
  /// sie in der Datenstruktur ab.
  /// </summary>
  /// <param name="lines">Zeilenanzahl</param>
  void RegisterLines(int lines)
  {
    if ((grammar != null) && (lines > grammar.maxLines))
    {
      grammar.maxLines = lines;
    }
  }

	void SynErr (int n) {
		if (errDist >= minErrDist) errors.SynErr(la.line, la.col, n);
		errDist = 0;
	}

	public void SemErr (string msg) {
		if (errDist >= minErrDist) errors.Error(t.line, t.col, msg);
		errDist = 0;
	}
	
	void Get () {
		for (;;) {
			t = la;
			la = scanner.Scan();
			if (la.kind <= maxT) { ++errDist; break; }
				if (la.kind == 25) {
				RegisterComment(t,la); 
				}
				if (la.kind == 26) {
				RegisterComment(t,la); 
				}
				if (la.kind == 27) {
				RegisterComment(t,la); 
				}
				if (la.kind == 28) {
				}

			la = t;
		}
	}
	
	void Expect (int n) {
		if (la.kind==n) Get(); else { SynErr(n); }
	}
	
	bool StartOf (int s) {
		return set[s, la.kind];
	}
	
	void ExpectWeak (int n, int follow) {
		if (la.kind == n) Get();
		else {
			SynErr(n);
			while (!StartOf(follow)) Get();
		}
	}
	
	bool WeakSeparator (int n, int syFol, int repFol) {
		bool[] s = new bool[maxT+1];
		if (la.kind == n) { Get(); return true; }
		else if (StartOf(repFol)) return false;
		else {
			for (int i=0; i <= maxT; i++) {
				s[i] = set[syFol, i] || set[repFol, i] || set[0, i];
			}
			SynErr(n);
			while (!s[la.kind]) Get();
			return StartOf(syFol);
		}
	}
	
	void Printy() {
		Node n, curr = null; 
		while (StartOf(1)) {
			Get();
		}
		Expect(4);
		Expect(1);
		grammar = new Grammar(t.val); 
		while (StartOf(2)) {
			Get();
		}
		Expect(5);
		grammar.meta = new Node(Node.Log,scanner.buffer.GetString(0,la.pos),null,la);
		RegisterLines(la.line - 1);
		
		Production(out grammar.sub);
		while (la.kind == 1) {
			Production(out n);
			n.Enqueue(grammar.sub,ref curr); 
		}
		Expect(6);
		Token start = t; 
		Expect(1);
		Expect(7);
		grammar.meta.next = new Node(Node.Log,scanner.buffer.GetString(start.pos,la.pos),start,la);
		RegisterLines(la.line - start.line);
		
	}

	void Production(out Node node) {
		Node n; 
		Expect(1);
		grammar.nts[t.val] = node = new Node(Node.Production,t); 
		if (la.kind == 20) {
			Attribute(out node.meta);
		}
		if (la.kind == 22) {
			Action(out n);
			if (node.meta == null) node.meta = n; else node.meta.next = n; 
		}
		ExpectWeak(8, 3);
		node.open = t; 
		Expression(out node.sub);
		ExpectWeak(7, 4);
		node.close = t; 
	}

	void Attribute(out Node node) {
		Token start = la; 
		Expect(20);
		while (StartOf(5)) {
			Get();
		}
		Expect(21);
		node = new Node(Node.Attribute,scanner.buffer.GetString(start.pos,t.pos + 1),start,t);
		RegisterLines(t.line - start.line);
		
	}

	void Action(out Node node) {
		Token start = la; 
		Expect(22);
		while (StartOf(6)) {
			Get();
		}
		Expect(23);
		node = new Node(Node.Action,scanner.buffer.GetString(start.pos,t.pos + 2),start,t);
		RegisterLines(t.line - start.line);
		
	}

	void Expression(out Node node) {
		Node n, curr = null; 
		Term(out node);
		while (WeakSeparator(9,7,8) ) {
			Token sep = t; 
			Term(out n);
			n.EnqueueAlt(node,ref curr); n.sep = sep; 
		}
	}

	void Term(out Node node) {
		Node n, curr = null; 
		node = null; 
		if (StartOf(9)) {
			if (la.kind == 19) {
				Resolver(out node);
			}
			Factor(out n);
			if (node == null) node = n; else n.Enqueue(node,ref curr); 
			while (StartOf(10)) {
				Factor(out n);
				n.Enqueue(node,ref curr); 
			}
		} else if (StartOf(11)) {
			node = new Node(Node.Epsilon); 
		} else SynErr(25);
		if (node == null) node = new Node(Node.Epsilon); 
	}

	void Resolver(out Node node) {
		Token start = la; 
		Expect(19);
		Expect(10);
		Condition();
		node = new Node(Node.Resolver,scanner.buffer.GetString(start.pos,t.pos + 1),start,t);
		RegisterLines(t.line - start.line);
		
	}

	void Factor(out Node node) {
		node = null; 
		switch (la.kind) {
		case 1: {
			Symbol(out node);
			break;
		}
		case 2: case 3: case 16: case 17: case 18: {
			Literal(out node);
			break;
		}
		case 22: {
			Action(out node);
			break;
		}
		case 10: {
			Get();
			node = new Node(Node.Group); node.open = t; 
			Expression(out node.sub);
			Expect(11);
			node.close = t;  
			break;
		}
		case 12: {
			Get();
			node = new Node(Node.Group); node.open = t; 
			Expression(out node.sub);
			Expect(13);
			node.close = t;  
			break;
		}
		case 14: {
			Get();
			node = new Node(Node.Group); node.open = t; 
			Expression(out node.sub);
			Expect(15);
			node.close = t;  
			break;
		}
		default: SynErr(26); break;
		}
		if (node == null) node = new Node(Node.Epsilon); 
	}

	void Symbol(out Node node) {
		node = new Node(Node.Symbol,la); 
		Expect(1);
		if (la.kind == 20) {
			Attribute(out node.meta);
		}
	}

	void Literal(out Node node) {
		node = new Node(Node.Literal,la); 
		if (la.kind == 2) {
			Get();
		} else if (la.kind == 3) {
			Get();
		} else if (la.kind == 16) {
			Get();
		} else if (la.kind == 17) {
			Get();
		} else if (la.kind == 18) {
			Get();
		} else SynErr(27);
	}

	void Condition() {
		while (StartOf(12)) {
			if (la.kind == 10) {
				Get();
				Condition();
			} else {
				Get();
			}
		}
		Expect(11);
	}



	public void Parse() {
		la = new Token();
		la.val = "";		
		Get();
		Printy();

    Expect(0);
	}
	
	bool[,] set = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x},
		{x,T,T,T, x,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,x},
		{x,T,T,T, T,x,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,x},
		{T,T,T,T, x,x,x,T, x,T,T,x, T,x,T,x, T,T,T,T, x,x,T,x, x,x},
		{T,T,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x},
		{x,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,x,T,T, T,x},
		{x,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,x, T,x},
		{x,T,T,T, x,x,x,T, x,T,T,T, T,T,T,T, T,T,T,T, x,x,T,x, x,x},
		{x,x,x,x, x,x,x,T, x,x,x,T, x,T,x,T, x,x,x,x, x,x,x,x, x,x},
		{x,T,T,T, x,x,x,x, x,x,T,x, T,x,T,x, T,T,T,T, x,x,T,x, x,x},
		{x,T,T,T, x,x,x,x, x,x,T,x, T,x,T,x, T,T,T,x, x,x,T,x, x,x},
		{x,x,x,x, x,x,x,T, x,T,x,T, x,T,x,T, x,x,x,x, x,x,x,x, x,x},
		{x,T,T,T, T,T,T,T, T,T,T,x, T,T,T,T, T,T,T,T, T,T,T,T, T,x}

	};
} // end Parser


public class Errors {
	public int count = 0;                                    // number of errors detected
  public string errMsgFormat = "-- line {0} col {1}: {2}"; // 0=line, 1=column, 2=text
	
	public void SynErr (int line, int col, int n) {
		string s;
		switch (n) {
			case 0: s = "EOF expected"; break;
			case 1: s = "ident expected"; break;
			case 2: s = "string expected"; break;
			case 3: s = "char expected"; break;
			case 4: s = "\"COMPILER\" expected"; break;
			case 5: s = "\"PRODUCTIONS\" expected"; break;
			case 6: s = "\"END\" expected"; break;
			case 7: s = "\".\" expected"; break;
			case 8: s = "\"=\" expected"; break;
			case 9: s = "\"|\" expected"; break;
			case 10: s = "\"(\" expected"; break;
			case 11: s = "\")\" expected"; break;
			case 12: s = "\"[\" expected"; break;
			case 13: s = "\"]\" expected"; break;
			case 14: s = "\"{\" expected"; break;
			case 15: s = "\"}\" expected"; break;
			case 16: s = "\"WEAK\" expected"; break;
			case 17: s = "\"SYNC\" expected"; break;
			case 18: s = "\"ANY\" expected"; break;
			case 19: s = "\"IF\" expected"; break;
			case 20: s = "\"<\" expected"; break;
			case 21: s = "\">\" expected"; break;
			case 22: s = "\"(.\" expected"; break;
			case 23: s = "\".)\" expected"; break;
			case 24: s = "??? expected"; break;
			case 25: s = "invalid Term"; break;
			case 26: s = "invalid Factor"; break;
			case 27: s = "invalid Literal"; break;

			default: s = "error " + n; break;
		}
		Console.WriteLine(errMsgFormat, line, col, s);
		count++;
	}

	public void SemErr (int line, int col, int n) {
		Console.WriteLine(errMsgFormat, line, col, ("error " + n));
		count++;
	}

	public void Error (int line, int col, string s) {
		Console.WriteLine(errMsgFormat, line, col, s);
		count++;
	}

	public void Exception (string s) {
		Console.WriteLine(s); 
		System.Environment.Exit(1);
	}
} // Errors

}