using Printy.Library;

using System;

namespace Printy.Tool {



public class Parser {
	const int _EOF = 0;
	const int _number = 1;
	const int maxT = 36;

	const bool T = true;
	const bool x = false;
	const int minErrDist = 2;
	
	public Scanner scanner;
	public Errors  errors;

	public Token t;    // last recognized token
	public Token la;   // lookahead token
	int errDist = minErrDist;

public Config config;
  public HyperConfig htmlConfig;

/*-------------------------------------------------------------------------*/



	public Parser(Scanner scanner) {
		this.scanner = scanner;
		errors = new Errors();
	}

	void SynErr (int n) {
		if (errDist >= minErrDist) errors.SynErr(la.line, la.col, n);
		errDist = 0;
	}

	public void SemErr (string msg) {
		if (errDist >= minErrDist) errors.Error(t.line, t.col, msg);
		errDist = 0;
	}
	
	void Get () {
		for (;;) {
			t = la;
			la = scanner.Scan();
			if (la.kind <= maxT) { ++errDist; break; }

			la = t;
		}
	}
	
	void Expect (int n) {
		if (la.kind==n) Get(); else { SynErr(n); }
	}
	
	bool StartOf (int s) {
		return set[s, la.kind];
	}
	
	void ExpectWeak (int n, int follow) {
		if (la.kind == n) Get();
		else {
			SynErr(n);
			while (!StartOf(follow)) Get();
		}
	}
	
	bool WeakSeparator (int n, int syFol, int repFol) {
		bool[] s = new bool[maxT+1];
		if (la.kind == n) { Get(); return true; }
		else if (StartOf(repFol)) return false;
		else {
			for (int i=0; i <= maxT; i++) {
				s[i] = set[syFol, i] || set[repFol, i] || set[0, i];
			}
			SynErr(n);
			while (!s[la.kind]) Get();
			return StartOf(syFol);
		}
	}
	
	void PrintyConfig() {
		while (StartOf(1)) {
			Rule();
			Expect(2);
		}
	}

	void Rule() {
		if (StartOf(2)) {
			ConfigRule();
		} else if (la.kind == 12 || la.kind == 13) {
			HyperConfigRule();
		} else SynErr(37);
	}

	void ConfigRule() {
		if (config == null) errors.Exception("Config not set."); 
		switch (la.kind) {
		case 3: {
			SpaceRule();
			break;
		}
		case 4: {
			BreakRule();
			break;
		}
		case 8: {
			OpenRule();
			break;
		}
		case 9: {
			CloseRule();
			break;
		}
		case 10: {
			HideRule();
			break;
		}
		case 11: {
			OverlapRule();
			break;
		}
		default: SynErr(38); break;
		}
	}

	void HyperConfigRule() {
		if (htmlConfig == null) errors.Exception("HyperConfig not set."); 
		if (la.kind == 12) {
			StyleRule();
		} else if (la.kind == 13) {
			ClassRule();
		} else SynErr(39);
	}

	void SpaceRule() {
		int symbol, follow; 
		Expect(3);
		SymbolAllIndent(out symbol);
		SymbolAll(out follow);
		Expect(1);
		config.PutSpace(symbol,follow,int.Parse(t.val)); 
	}

	void BreakRule() {
		int symbol, follow; 
		Expect(4);
		SymbolAll(out symbol);
		SymbolAll(out follow);
		if (la.kind == 1) {
			Get();
			config.PutBreak(symbol,follow,int.Parse(t.val)); 
		} else if (la.kind == 5) {
			Get();
			config.PutBreak(symbol,follow,Config.BREAK_DEFAULT); 
		} else if (la.kind == 6) {
			Get();
			config.PutBreak(symbol,follow,Config.BREAK_ALWAYS); 
		} else if (la.kind == 7) {
			Get();
			config.PutBreak(symbol,follow,Config.BREAK_NEVER); 
		} else SynErr(40);
	}

	void OpenRule() {
		Expect(8);
		if (la.kind == 14) {
			BitSet(out config.openSet);
		} else if (StartOf(3)) {
			Bit(ref config.openSet);
		} else SynErr(41);
	}

	void CloseRule() {
		Expect(9);
		if (la.kind == 14) {
			BitSet(out config.closeSet);
		} else if (StartOf(3)) {
			Bit(ref config.closeSet);
		} else SynErr(42);
	}

	void HideRule() {
		Expect(10);
		if (la.kind == 14) {
			BitSet(out config.hideSet);
		} else if (StartOf(3)) {
			Bit(ref config.hideSet);
		} else SynErr(43);
	}

	void OverlapRule() {
		Expect(11);
		if (la.kind == 14) {
			BitSet(out config.overlapSet);
		} else if (StartOf(3)) {
			Bit(ref config.overlapSet);
		} else SynErr(44);
	}

	void StyleRule() {
		Expect(12);
		if (la.kind == 14) {
			BitSet(out htmlConfig.styleSet);
		} else if (StartOf(3)) {
			Bit(ref htmlConfig.styleSet);
		} else SynErr(45);
	}

	void ClassRule() {
		string text, className; int symbol = Config.ALL; 
		Expect(13);
		Text(out className);
		if (StartOf(3)) {
			SymbolAll(out symbol);
		}
		Text(out text);
		htmlConfig.PutStyleClass(text,symbol,className); 
		while (la.kind == 34) {
			Text(out text);
			htmlConfig.PutStyleClass(text,symbol,className); 
		}
	}

	void SymbolAllIndent(out int type) {
		type = 0; 
		if (la.kind == 17) {
			Get();
			type = Config.INDENT; 
		} else if (StartOf(3)) {
			SymbolAll(out type);
		} else SynErr(46);
	}

	void SymbolAll(out int type) {
		type = 0; 
		if (la.kind == 16) {
			Get();
			type = Config.ALL; 
		} else if (StartOf(4)) {
			Symbol(out type);
		} else SynErr(47);
	}

	void BitSet(out BitSet vector) {
		int symbol; 
		vector = new BitSet(); 
		Expect(14);
		while (StartOf(4)) {
			Symbol(out symbol);
			vector += symbol; 
		}
		Expect(15);
	}

	void Bit(ref BitSet vector) {
		int symbol; 
		SymbolAll(out symbol);
		if (symbol == Config.ALL) vector |= true; else vector += symbol; 
	}

	void Text(out string text) {
		Token start = la; 
		Expect(34);
		while (StartOf(5)) {
			Get();
		}
		Expect(35);
		text = scanner.buffer.GetString(start.pos + 1,t.pos); 
	}

	void Symbol(out int type) {
		type = 0; 
		switch (la.kind) {
		case 18: {
			Get();
			type = Config.Production; 
			break;
		}
		case 19: {
			Get();
			type = Config.Symbol; 
			break;
		}
		case 20: {
			Get();
			type = Config.Literal; 
			break;
		}
		case 21: {
			Get();
			type = Config.Open; 
			break;
		}
		case 22: {
			Get();
			type = Config.Close; 
			break;
		}
		case 23: {
			Get();
			type = Config.Assign; 
			break;
		}
		case 24: {
			Get();
			type = Config.Separator; 
			break;
		}
		case 25: {
			Get();
			type = Config.Delimiter; 
			break;
		}
		case 26: {
			Get();
			type = Config.InnerComment; 
			break;
		}
		case 27: {
			Get();
			type = Config.OuterComment; 
			break;
		}
		case 28: {
			Get();
			type = Config.Attribute; 
			break;
		}
		case 29: {
			Get();
			type = Config.Resolver; 
			break;
		}
		case 30: {
			Get();
			type = Config.Prolog; 
			break;
		}
		case 31: {
			Get();
			type = Config.Epilog; 
			break;
		}
		case 32: {
			Get();
			type = Config.Action; 
			break;
		}
		case 33: {
			Get();
			type = Config.LineNumber; 
			break;
		}
		default: SynErr(48); break;
		}
	}



	public void Parse() {
		la = new Token();
		la.val = "";		
		Get();
		PrintyConfig();

    Expect(0);
	}
	
	bool[,] set = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x},
		{x,x,x,T, T,x,x,x, T,T,T,T, T,T,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x},
		{x,x,x,T, T,x,x,x, T,T,T,T, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x},
		{x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,x,x, x,x},
		{x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,x,x, x,x},
		{x,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,x, T,x}

	};
} // end Parser


public class Errors {
	public int count = 0;                                    // number of errors detected
  public string errMsgFormat = "-- line {0} col {1}: {2}"; // 0=line, 1=column, 2=text
	
	public void SynErr (int line, int col, int n) {
		string s;
		switch (n) {
			case 0: s = "EOF expected"; break;
			case 1: s = "number expected"; break;
			case 2: s = "\";\" expected"; break;
			case 3: s = "\"space\" expected"; break;
			case 4: s = "\"break\" expected"; break;
			case 5: s = "\"default\" expected"; break;
			case 6: s = "\"always\" expected"; break;
			case 7: s = "\"never\" expected"; break;
			case 8: s = "\"open\" expected"; break;
			case 9: s = "\"close\" expected"; break;
			case 10: s = "\"hide\" expected"; break;
			case 11: s = "\"overlap\" expected"; break;
			case 12: s = "\"style\" expected"; break;
			case 13: s = "\"class\" expected"; break;
			case 14: s = "\"[\" expected"; break;
			case 15: s = "\"]\" expected"; break;
			case 16: s = "\"ALL\" expected"; break;
			case 17: s = "\"INDENT\" expected"; break;
			case 18: s = "\"Production\" expected"; break;
			case 19: s = "\"Symbol\" expected"; break;
			case 20: s = "\"Literal\" expected"; break;
			case 21: s = "\"Open\" expected"; break;
			case 22: s = "\"Close\" expected"; break;
			case 23: s = "\"Assign\" expected"; break;
			case 24: s = "\"Separator\" expected"; break;
			case 25: s = "\"Delimiter\" expected"; break;
			case 26: s = "\"InnerComment\" expected"; break;
			case 27: s = "\"OuterComment\" expected"; break;
			case 28: s = "\"Attribute\" expected"; break;
			case 29: s = "\"Resolver\" expected"; break;
			case 30: s = "\"Prolog\" expected"; break;
			case 31: s = "\"Epilog\" expected"; break;
			case 32: s = "\"Action\" expected"; break;
			case 33: s = "\"LineNumber\" expected"; break;
			case 34: s = "\"<\" expected"; break;
			case 35: s = "\">\" expected"; break;
			case 36: s = "??? expected"; break;
			case 37: s = "invalid Rule"; break;
			case 38: s = "invalid ConfigRule"; break;
			case 39: s = "invalid HyperConfigRule"; break;
			case 40: s = "invalid BreakRule"; break;
			case 41: s = "invalid OpenRule"; break;
			case 42: s = "invalid CloseRule"; break;
			case 43: s = "invalid HideRule"; break;
			case 44: s = "invalid OverlapRule"; break;
			case 45: s = "invalid StyleRule"; break;
			case 46: s = "invalid SymbolAllIndent"; break;
			case 47: s = "invalid SymbolAll"; break;
			case 48: s = "invalid Symbol"; break;

			default: s = "error " + n; break;
		}
		Console.WriteLine(errMsgFormat, line, col, s);
		count++;
	}

	public void SemErr (int line, int col, int n) {
		Console.WriteLine(errMsgFormat, line, col, ("error " + n));
		count++;
	}

	public void Error (int line, int col, string s) {
		Console.WriteLine(errMsgFormat, line, col, s);
		count++;
	}

	public void Exception (string s) {
		Console.WriteLine(s); 
		System.Environment.Exit(1);
	}
} // Errors

}